/*
 * This test program illustrates how a process may use the
 * Linux x86 Performance-Monitoring Counters interface to
 * monitor its own execution.
 *
 * Copyright (C) 1999  Mikael Pettersson
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "libperfctr.h"

unsigned char events_format;

const char *events_format_string(unsigned events_format)
{
    switch( events_format ) {
      case PERFCTR_EVENTS_INTEL_P5:
	return "Intel Pentium";
      case PERFCTR_EVENTS_INTEL_P5MMX:
	return "Intel Pentium MMX";
      case PERFCTR_EVENTS_INTEL_P6:
	return "Intel Pentium Pro";
      case PERFCTR_EVENTS_INTEL_PII:
	return "Intel Pentium II";
      case PERFCTR_EVENTS_CYRIX_MII:
	return "Cyrix MII";
      default:
	return "?";
    }
}

unsigned event_num_insns(void)
{
    switch( events_format ) {
      case PERFCTR_EVENTS_INTEL_P5:
      case PERFCTR_EVENTS_INTEL_P5MMX:
      case PERFCTR_EVENTS_CYRIX_MII:
	/* event 0x16, count at CPL 3 */
	return 0x16 | (2 << 6);
      case PERFCTR_EVENTS_INTEL_P6:
      case PERFCTR_EVENTS_INTEL_PII:
	/* event 0xc0, count at CPL > 0, ENable */
	return 0xc0 | (1 << 16) | (1 << 22);
      case PERFCTR_EVENTS_WINCHIP_C6:
	return 0x02;
      case PERFCTR_EVENTS_WINCHIP_2:
	return 0x16;
      default:
	fprintf(stderr, "events_format %u not recognized\n", events_format);
	exit(1);
    }
}

void do_init(void)
{
    struct perfctr_info info;

    if( perfctr_info(&info) < 0 ) {
	perror("perfctr_info");
	exit(1);
    }
    events_format = info.events_format;
    printf("\nPerfCtr Info:\n");
    printf("version_major\t\t%u\n", info.version_major);
    printf("version_minor\t\t%u\n", info.version_minor);
    printf("version_pre\t\t%u\n", info.version_pre);
    printf("events_format\t\t%u (%s)\n",
	   info.events_format,
	   events_format_string(info.events_format));
    printf("have_rdpmc\t\t%u\n", info.have_rdpmc);
    printf("have_rdtsc\t\t%u\n", info.have_rdtsc);
}

void do_read(void)
{
    struct perfctr_state state;

    if( perfctr_read_self(&state) < 0 ) {
	 perror("perfctr_read_self");
	 exit(1);
    }
    printf("\nCurrent Sample:\n");
    printf("status\t\t\t%d\n", state.status);
    printf("control.evntsel[0]\t0x%08X\n", state.control.evntsel[0]);
    printf("control.evntsel[1]\t0x%08X\n", state.control.evntsel[1]);
    printf("tsc_start\t\t0x%016llX\n", state.counters.tsc_start);
    printf("tsc\t\t\t0x%016llX\n", state.counters.tsc);
    printf("pmc[0]\t\t\t0x%016llX\n", state.counters.pmc[0]);
    printf("pmc[1]\t\t\t0x%016llX\n", state.counters.pmc[1]);
}

void do_enable(unsigned evntsel0)
{
    struct perfctr_control control;

    control.evntsel[0] = evntsel0;
    control.evntsel[1] = 0;
    if( perfctr_control_self(&control) < 0 ) {
	perror("perfctr_control_self");
	exit(1);
    }
}

unsigned fac(unsigned n)
{
    return (n < 2) ? 1 : n * fac(n-1);
}

void do_fac(unsigned n)
{
    printf("\nfac(%u) == %u\n", n, fac(n));
}

int main(void)
{
    do_init();
    do_read();
    do_enable(event_num_insns());
    do_fac(15);
    do_read();
    return 0;
}
